@@ -1,13 +1,37 @@
-#include <sys/types.h>
-#include <sys/socket.h>
-#include <netinet/in.h>
+#include "btpd.h"
 
 #include <ctype.h>
-#include <math.h>
-#include <string.h>
-#include <unistd.h>
 
-#include "btpd.h"
+struct meta_peer *
+mp_create(void)
+{
+    return btpd_calloc(1, sizeof(struct meta_peer));
+}
+
+void
+mp_kill(struct meta_peer *mp)
+{
+    free(mp);
+}
+
+void
+mp_hold(struct meta_peer *mp)
+{
+    mp->refs++;
+}
+
+void
+mp_drop(struct meta_peer *mp, struct net *n)
+{
+    assert(mp->refs > 0);
+    mp->refs--;
+    if (mp->refs == 0) {
+        assert(mp->p == NULL);
+        if (mp->flags & PF_ATTACHED)
+            assert(mptbl_remove(n->mptbl, mp->id) == mp);
+        mp_kill(mp);
+    }
+}
 
 void
 peer_kill(struct peer *p)
@@ -16,7 +40,7 @@ peer_kill(struct peer *p)
 
     btpd_log(BTPD_L_CONN, "killed peer %p\n", p);
 
-    if (p->flags & PF_ATTACHED) {
+    if (p->mp->flags & PF_ATTACHED) {
         BTPDQ_REMOVE(&p->n->peers, p, p_entry);
         p->n->npeers--;
         if (p->n->active) {
@@ -25,13 +49,12 @@ peer_kill(struct peer *p)
         }
     } else
         BTPDQ_REMOVE(&net_unattached, p, p_entry);
-    if (p->flags & PF_ON_READQ)
+    if (p->mp->flags & PF_ON_READQ)
         BTPDQ_REMOVE(&net_bw_readq, p, rq_entry);
-    if (p->flags & PF_ON_WRITEQ)
+    if (p->mp->flags & PF_ON_WRITEQ)
         BTPDQ_REMOVE(&net_bw_writeq, p, wq_entry);
 
-    btpd_ev_del(&p->in_ev);
-    btpd_ev_del(&p->out_ev);
+    btpd_ev_del(&p->ioev);
     close(p->sd);
 
     nl = BTPDQ_FIRST(&p->outq);
@@ -42,10 +65,14 @@ peer_kill(struct peer *p)
         nl = next;
     }
 
+    p->mp->p = NULL;
+    mp_drop(p->mp, p->n);
     if (p->in.buf != NULL)
         free(p->in.buf);
     if (p->piece_field != NULL)
         free(p->piece_field);
+    if (p->bad_field != NULL)
+        free(p->bad_field);
     free(p);
     net_npeers--;
 }
@@ -66,7 +93,7 @@ peer_send(struct peer *p, struct net_buf *nb)
 
     if (BTPDQ_EMPTY(&p->outq)) {
         assert(p->outq_off == 0);
-        btpd_ev_add(&p->out_ev, NULL);
+        btpd_ev_enable(&p->ioev, EV_WRITE);
         p->t_wantwrite = btpd_seconds;
     }
     BTPDQ_INSERT_TAIL(&p->outq, nl, entry);
@@ -91,11 +118,11 @@ peer_unsend(struct peer *p, struct nb_link *nl)
         nb_drop(nl->nb);
         free(nl);
         if (BTPDQ_EMPTY(&p->outq)) {
-            if (p->flags & PF_ON_WRITEQ) {
+            if (p->mp->flags & PF_ON_WRITEQ) {
                 BTPDQ_REMOVE(&net_bw_writeq, p, wq_entry);
-                p->flags &= ~PF_ON_WRITEQ;
+                p->mp->flags &= ~PF_ON_WRITEQ;
             } else
-                btpd_ev_del(&p->out_ev);
+                btpd_ev_disable(&p->ioev, EV_WRITE);
         }
         return 1;
     } else
@@ -114,7 +141,7 @@ peer_sent(struct peer *p, struct net_buf *nb)
         break;
     case NB_UNCHOKE:
         btpd_log(BTPD_L_MSG, "sent unchoke to %p\n", p);
-        p->flags &= ~PF_NO_REQUESTS;
+        p->mp->flags &= ~PF_NO_REQUESTS;
         break;
     case NB_INTEREST:
         btpd_log(BTPD_L_MSG, "sent interest to %p\n", p);
@@ -207,7 +234,7 @@ peer_cancel(struct peer *p, struct block_request *req, struct net_buf *nb)
 void
 peer_unchoke(struct peer *p)
 {
-    p->flags &= ~PF_I_CHOKE;
+    p->mp->flags &= ~PF_I_CHOKE;
     peer_send(p, nb_create_unchoke());
 }
 
@@ -226,18 +253,23 @@ peer_choke(struct peer *p)
         nl = next;
     }
 
-    p->flags |= PF_I_CHOKE;
+    p->mp->flags |= PF_I_CHOKE;
     peer_send(p, nb_create_choke());
 }
 
 void
 peer_want(struct peer *p, uint32_t index)
 {
+    if (!has_bit(p->piece_field, index) || peer_has_bad(p, index))
+        return;
     assert(p->nwant < p->npieces);
     p->nwant++;
     if (p->nwant == 1) {
+        p->mp->flags |= PF_I_WANT;
+        if (p->mp->flags & PF_SUSPECT)
+            return;
         if (p->nreqs_out == 0) {
-            assert((p->flags & PF_DO_UNWANT) == 0);
+            assert((p->mp->flags & PF_DO_UNWANT) == 0);
             int unsent = 0;
             struct nb_link *nl = BTPDQ_LAST(&p->outq, nb_tq);
             if (nl != NULL && nl->nb->type == NB_UNINTEREST)
@@ -245,25 +277,28 @@ peer_want(struct peer *p, uint32_t index)
             if (!unsent)
                 peer_send(p, nb_create_interest());
         } else {
-            assert((p->flags & PF_DO_UNWANT) != 0);
-            p->flags &= ~PF_DO_UNWANT;
+            assert((p->mp->flags & PF_DO_UNWANT) != 0);
+            p->mp->flags &= ~PF_DO_UNWANT;
         }
-        p->flags |= PF_I_WANT;
     }
 }
 
 void
 peer_unwant(struct peer *p, uint32_t index)
 {
+    if (!has_bit(p->piece_field, index) || peer_has_bad(p, index))
+        return;
     assert(p->nwant > 0);
     p->nwant--;
     if (p->nwant == 0) {
-        p->flags &= ~PF_I_WANT;
+        p->mp->flags &= ~PF_I_WANT;
+        if (p->mp->flags & PF_SUSPECT)
+            return;
         p->t_nointerest = btpd_seconds;
         if (p->nreqs_out == 0)
             peer_send(p, nb_create_uninterest());
         else
-            p->flags |= PF_DO_UNWANT;
+            p->mp->flags |= PF_DO_UNWANT;
     }
 }
 
@@ -272,8 +307,12 @@ peer_create_common(int sd)
 {
     struct peer *p = btpd_calloc(1, sizeof(*p));
 
+    p->mp = mp_create();
+    mp_hold(p->mp);
+    p->mp->p = p;
+
     p->sd = sd;
-    p->flags = PF_I_CHOKE | PF_P_CHOKE;
+    p->mp->flags = PF_I_CHOKE | PF_P_CHOKE;
     p->t_created = btpd_seconds;
     p->t_lastwrite = btpd_seconds;
     p->t_nointerest = btpd_seconds;
@@ -282,9 +321,7 @@ peer_create_common(int sd)
 
     peer_set_in_state(p, SHAKE_PSTR, 28);
 
-    event_set(&p->out_ev, p->sd, EV_WRITE, net_write_cb, p);
-    event_set(&p->in_ev, p->sd, EV_READ, net_read_cb, p);
-    btpd_ev_add(&p->in_ev, NULL);
+    btpd_ev_new(&p->ioev, p->sd, EV_READ, net_io_cb, p);
 
     BTPDQ_INSERT_TAIL(&net_unattached, p, p_entry);
     net_npeers++;
@@ -295,7 +332,7 @@ void
 peer_create_in(int sd)
 {
     struct peer *p = peer_create_common(sd);
-    p->flags |= PF_INCOMING;
+    p->mp->flags |= PF_INCOMING;
 }
 
 void
@@ -305,7 +342,7 @@ peer_create_out(struct net *n, const uint8_t *id,
     int sd;
     struct peer *p;
 
-    if (net_connect(ip, port, &sd) != 0)
+    if (net_connect_name(ip, port, &sd) != 0)
         return;
 
     p = peer_create_common(sd);
@@ -314,17 +351,36 @@ peer_create_out(struct net *n, const uint8_t *id,
 }
 
 void
-peer_create_out_compact(struct net *n, const char *compact)
+peer_create_out_compact(struct net *n, int family, const char *compact)
 {
     int sd;
     struct peer *p;
-    struct sockaddr_in addr;
-
-    addr.sin_family = AF_INET;
-    bcopy(compact, &addr.sin_addr.s_addr, 4);
-    bcopy(compact + 4, &addr.sin_port, 2);
-
-    if (net_connect2((struct sockaddr *)&addr, sizeof(addr), &sd) != 0)
+    struct sockaddr_storage addr;
+    struct sockaddr_in *a4;
+    struct sockaddr_in6 *a6;
+    
+    switch (family) {
+    case AF_INET:
+        if (!net_ipv4)
+            return;
+        a4 = (struct sockaddr_in *)&addr;
+        a4->sin_family = AF_INET;
+        bcopy(compact, &a4->sin_addr.s_addr, 4);
+        bcopy(compact + 4, &a4->sin_port, 2);
+        break;
+    case AF_INET6:
+        if (!net_ipv6)
+            return;
+        a6 = (struct sockaddr_in6 *)&addr;
+        a6->sin6_family = AF_INET6;
+        bcopy(compact, &a6->sin6_addr, 16);
+        bcopy(compact + 16, &a6->sin6_port, 2);
+        break;
+    default:
+        abort();
+    }
+    if (net_connect_addr(family, (struct sockaddr *)&addr,
+            sizeof(addr), &sd) != 0)
         return;
 
     p = peer_create_common(sd);
@@ -335,9 +391,9 @@ peer_create_out_compact(struct net *n, const char *compact)
 void
 peer_on_no_reqs(struct peer *p)
 {
-    if ((p->flags & PF_DO_UNWANT) != 0) {
+    if ((p->mp->flags & PF_DO_UNWANT) != 0) {
         assert(p->nwant == 0);
-        p->flags &= ~PF_DO_UNWANT;
+        p->mp->flags &= ~PF_DO_UNWANT;
         peer_send(p, nb_create_uninterest());
     }
 }
@@ -353,8 +409,8 @@ peer_on_shake(struct peer *p)
 {
     uint8_t printid[21];
     int i;
-    for (i = 0; i < 20 && isprint(p->id[i]); i++)
-        printid[i] = p->id[i];
+    for (i = 0; i < 20 && isprint(p->mp->id[i]); i++)
+        printid[i] = p->mp->id[i];
     printid[i] = '\0';
     btpd_log(BTPD_L_MSG, "received shake(%s) from %p\n", printid, p);
     p->piece_field = btpd_calloc(1, (int)ceil(p->n->tp->npieces / 8.0));
@@ -368,9 +424,10 @@ peer_on_shake(struct peer *p)
         }
     }
 
+    mptbl_insert(p->n->mptbl, p->mp);
     BTPDQ_REMOVE(&net_unattached, p, p_entry);
     BTPDQ_INSERT_HEAD(&p->n->peers, p, p_entry);
-    p->flags |= PF_ATTACHED;
+    p->mp->flags |= PF_ATTACHED;
     p->n->npeers++;
 
     ul_on_new_peer(p);
@@ -381,10 +438,10 @@ void
 peer_on_choke(struct peer *p)
 {
     btpd_log(BTPD_L_MSG, "received choke from %p\n", p);
-    if ((p->flags & PF_P_CHOKE) != 0)
+    if ((p->mp->flags & PF_P_CHOKE) != 0)
         return;
     else {
-        p->flags |= PF_P_CHOKE;
+        p->mp->flags |= PF_P_CHOKE;
         dl_on_choke(p);
         struct nb_link *nl = BTPDQ_FIRST(&p->outq);
         while (nl != NULL) {
@@ -400,10 +457,10 @@ void
 peer_on_unchoke(struct peer *p)
 {
     btpd_log(BTPD_L_MSG, "received unchoke from %p\n", p);
-    if ((p->flags & PF_P_CHOKE) == 0)
+    if ((p->mp->flags & PF_P_CHOKE) == 0)
         return;
     else {
-        p->flags &= ~PF_P_CHOKE;
+        p->mp->flags &= ~PF_P_CHOKE;
         dl_on_unchoke(p);
     }
 }
@@ -412,10 +469,10 @@ void
 peer_on_interest(struct peer *p)
 {
     btpd_log(BTPD_L_MSG, "received interest from %p\n", p);
-    if ((p->flags & PF_P_WANT) != 0)
+    if ((p->mp->flags & PF_P_WANT) != 0)
         return;
     else {
-        p->flags |= PF_P_WANT;
+        p->mp->flags |= PF_P_WANT;
         ul_on_interest(p);
     }
 }
@@ -424,10 +481,10 @@ void
 peer_on_uninterest(struct peer *p)
 {
     btpd_log(BTPD_L_MSG, "received uninterest from %p\n", p);
-    if ((p->flags & PF_P_WANT) == 0)
+    if ((p->mp->flags & PF_P_WANT) == 0)
         return;
     else {
-        p->flags &= ~PF_P_WANT;
+        p->mp->flags &= ~PF_P_WANT;
         p->t_nointerest = btpd_seconds;
         ul_on_uninterest(p);
     }
@@ -488,14 +545,14 @@ peer_on_request(struct peer *p, uint32_t index, uint32_t begin,
 {
     btpd_log(BTPD_L_MSG, "received request(%u,%u,%u) from %p\n",
         index, begin, length, p);
-    if ((p->flags & PF_NO_REQUESTS) == 0) {
+    if ((p->mp->flags & PF_NO_REQUESTS) == 0) {
         peer_send(p, nb_create_piece(index, begin, length));
         peer_send(p, nb_create_torrentdata());
         p->npiece_msgs++;
         if (p->npiece_msgs >= MAXPIECEMSGS) {
             peer_send(p, nb_create_choke());
             peer_send(p, nb_create_unchoke());
-            p->flags |= PF_NO_REQUESTS;
+            p->mp->flags |= PF_NO_REQUESTS;
         }
     }
 }
@@ -522,7 +579,9 @@ peer_on_cancel(struct peer *p, uint32_t index, uint32_t begin,
 void
 peer_on_tick(struct peer *p)
 {
-    if (p->flags & PF_ATTACHED) {
+    if (p->mp->flags & PF_BANNED)
+        goto kill;
+    if (p->mp->flags & PF_ATTACHED) {
         if (BTPDQ_EMPTY(&p->outq)) {
             if (btpd_seconds - p->t_lastwrite >= 120)
                 peer_keepalive(p);
@@ -530,7 +589,7 @@ peer_on_tick(struct peer *p)
             btpd_log(BTPD_L_CONN, "write attempt timed out.\n");
             goto kill;
         }
-        if ((cm_full(p->n->tp) && !(p->flags & PF_P_WANT) &&
+        if ((cm_full(p->n->tp) && !(p->mp->flags & PF_P_WANT) &&
                 btpd_seconds - p->t_nointerest >= 600)) {
             btpd_log(BTPD_L_CONN, "no interest for 10 minutes.\n");
             goto kill;
@@ -544,10 +603,56 @@ kill:
     peer_kill(p);
 }
 
+void
+peer_bad_piece(struct peer *p, uint32_t index)
+{
+    if (p->npcs_bad == 0) {
+        assert(p->bad_field == NULL);
+        p->bad_field = btpd_calloc(ceil(p->n->tp->npieces / 8.0), 1);
+    }
+    assert(!has_bit(p->bad_field, index));
+    set_bit(p->bad_field, index);
+    p->npcs_bad++;
+    p->suspicion++;
+    if (p->suspicion == 3) {
+        btpd_log(BTPD_L_BAD, "suspect peer %p.\n", p);
+        p->mp->flags |= PF_SUSPECT;
+        if (p->nwant > 0) {
+            p->mp->flags &= ~PF_DO_UNWANT;
+            peer_send(p, nb_create_uninterest());
+        }
+    }
+}
+
+void
+peer_good_piece(struct peer *p, uint32_t index)
+{
+    if (peer_has_bad(p, index)) {
+        assert(p->npcs_bad > 0);
+        p->npcs_bad--;
+        if (p->npcs_bad == 0) {
+            free(p->bad_field);
+            p->bad_field = NULL;
+        } else
+            clear_bit(p->bad_field, index);
+    }
+    p->suspicion = 0;
+    if (p->mp->flags & PF_SUSPECT) {
+        btpd_log(BTPD_L_BAD, "unsuspect peer %p.\n", p);
+        p->mp->flags &= ~PF_SUSPECT;
+        if (p->nwant > 0) {
+            assert(p->mp->flags & PF_I_WANT);
+            peer_send(p, nb_create_interest());
+        }
+        if (peer_leech_ok(p))
+            dl_on_download(p);
+    }
+}
+
 int
 peer_chokes(struct peer *p)
 {
-    return p->flags & PF_P_CHOKE;
+    return p->mp->flags & PF_P_CHOKE;
 }
 
 int
@@ -556,6 +661,12 @@ peer_has(struct peer *p, uint32_t index)
     return has_bit(p->piece_field, index);
 }
 
+int
+peer_has_bad(struct peer *p, uint32_t index)
+{
+    return p->bad_field != NULL && has_bit(p->bad_field, index);
+}
+
 int
 peer_laden(struct peer *p)
 {
@@ -565,13 +676,14 @@ peer_laden(struct peer *p)
 int
 peer_wanted(struct peer *p)
 {
-    return (p->flags & PF_I_WANT) == PF_I_WANT;
+    return (p->mp->flags & PF_I_WANT) == PF_I_WANT;
 }
 
 int
 peer_leech_ok(struct peer *p)
 {
-    return (p->flags & (PF_I_WANT|PF_P_CHOKE)) == PF_I_WANT;
+    return (p->mp->flags & (PF_BANNED|PF_SUSPECT|PF_I_WANT|PF_P_CHOKE))
+        == PF_I_WANT && !peer_laden(p);
 }
 
 int
@@ -583,7 +695,7 @@ peer_active_down(struct peer *p)
 int
 peer_active_up(struct peer *p)
 {
-    return (p->flags & (PF_P_WANT|PF_I_CHOKE)) == PF_P_WANT
+    return (p->mp->flags & (PF_P_WANT|PF_I_CHOKE)) == PF_P_WANT
         || p->npiece_msgs > 0;
 }
 
@@ -592,3 +704,9 @@ peer_full(struct peer *p)
 {
     return p->npieces == p->n->tp->npieces;
 }
+
+int
+peer_requestable(struct peer *p, uint32_t index)
+{
+    return peer_has(p, index) && !peer_has_bad(p, index);
+}
